"""Module with functions for 'project_utils' subpackage."""

from __future__ import annotations
from typing import Sequence, Any
import os
import sys

from typing_extensions import Literal

from .. import tests
from .. import venvs
from mypythontools.config import ConfigBase, MyProperty
from mypythontools.misc import GLOBAL_VARS, EMOJIS, print_progress
from mypythontools.paths import PathLike
from mypythontools.types import validate_sequence
from ..deploy import deploy_to_pypi
from .project_utils_functions import (
    get_version,
    git_push,
    reformat_with_black,
    set_version,
    docs_regenerate,
)

# Lazy loaded
# from git import Repo

from mypythontools_cicd.project_paths import PROJECT_PATHS


class PipelineConfig(ConfigBase):
    """Allow to setup CICD pipeline."""

    @MyProperty
    @staticmethod
    def do_only() -> Literal[
        None, "reformat", "test", "docs", "sync_requirements", "commit_and_push_git", "deploy"
    ]:
        """Run just single function from pipeline, ignore the others.

        Type:
            Literal[
                None, "reformat", "test", "docs", "sync_requirements", "commit_and_push_git", "deploy"
            ]

        Default:
            None

        Reason for why to call it form here and not directly is to be able to use sys args or single command
        line entrypoint.
        """
        return None

    @MyProperty
    @staticmethod
    def reformat() -> bool:
        """Reformat all python files with black. Setup parameters in pyproject.toml.

        Type:
            bool

        Default:
            True.
        """
        return True

    @MyProperty
    @staticmethod
    def test() -> bool:
        """Run pytest tests.

        Type:
            bool

        Default:
            True
        """
        return True

    @MyProperty
    @staticmethod
    def test_options() -> None | dict:
        """Check tests module and function run_tests for what parameters you can use.

        Type:
            None | dict

        Default:
            None

        For example:
            >>> {"virtualenvs": ["venv/37", "venv/310], "test_coverage": True, "verbose": False}
        """
        return None

    @MyProperty
    @staticmethod
    def version() -> None | str:
        """Overwrite __version__ in __init__.py.

        Type:
            str

        Default:
            'increment'.

        Version has to be in format like '1.0.3' three digits and two dots. If 'None', nothing will happen. If
        'increment', than it will be updated by 0.0.1..
        """
        return "increment"

    @MyProperty
    @staticmethod
    def docs() -> bool:
        """Whether generate sphinx apidoc and generate rst files for documentation. Some files in docs source
        can be deleted - check `docs` docstrings for details.

        Type:
            bool

        Default:
            True
        """
        return True

    @MyProperty
    @staticmethod
    def sync_requirements() -> None | Literal["infer"] | PathLike | Sequence[PathLike]:
        """Check requirements.txt and update all the libraries.

        Type:
            None | Literal["infer"] | PathLike | Sequence[PathLike]

        Default:
            None

        You can use path to requirements, list of paths or bool value. If True, then path is inferred.
        """
        return None

    @MyProperty
    @staticmethod
    def commit_and_push_git() -> bool:
        """Whether push to github or not.

        Type:
            bool

        Default:
            True
        """
        return True

    @MyProperty
    @staticmethod
    def commit_message() -> str:
        """Commit message.

        Type:
            str

        Default:
            'New commit'
        """
        return "New commit"

    @MyProperty
    @staticmethod
    def tag() -> str:
        """Tag. E.g 'v1.1.2'. If '__version__', get the version.

        Type:
            str

        Default:
            '__version__'
        """
        return "__version__"

    @MyProperty
    @staticmethod
    def tag_message() -> str:
        """Tag message.

        Type:
            bool

        Default:
            'New version'
        """
        return "New version"

    @MyProperty
    @staticmethod
    def deploy() -> bool:
        """Deploy to PYPI.

        `TWINE_USERNAME` and `TWINE_PASSWORD` are used for authorization.

        Type:
            bool

        Default:
            False
        """
        return False

    @MyProperty
    @staticmethod
    def allowed_branches() -> None | Sequence[str]:
        """Pipeline runs only on defined branches.

        Type:
            None | Sequence[str]

        Default:
            ["master", "main"]
        """
        return ["master", "main"]

    @MyProperty
    @staticmethod
    def verbosity() -> Literal[0, 1, 2]:
        """Pipeline runs only on defined branches.

        Type:
            Literal[0, 1, 2]

        Default:
            1
        """
        return 1


def project_utils_pipeline(
    config: None | PipelineConfig = None,
    do_only: Literal[
        None,
        "reformat",
        "test",
        "docs",
        "sync_requirements",
        "commit_and_push_git",
        "deploy",
    ] = None,
    reformat: bool = True,
    test: bool = True,
    test_options: None | dict[str, Sequence[PathLike]] | dict[str, Any] = None,
    version: None | str = "increment",
    docs: bool = True,
    sync_requirements: None | Literal["infer"] | PathLike | Sequence[PathLike] = None,
    commit_and_push_git: bool = True,
    commit_message: str = "New commit",
    tag: str = "__version__",
    tag_message: str = "New version",
    deploy: bool = False,
    allowed_branches: None | Sequence[str] = ("master", "main"),
    verbosity: Literal[0, 1, 2] = 1,
) -> None:
    """Run pipeline for pushing and deploying app.

    Can run tests, generate rst files for sphinx docs, push to github and deploy to pypi. All params can be
    configured not only with function params, but also from command line with params and therefore callable
    from terminal and optimal to run from IDE (for example with creating simple VS Code task).

    Some function suppose some project structure (where are the docs, where is __init__.py etc.).
    If you are issuing some error, try functions directly, find necessary paths in parameters
    and set paths that are necessary in paths module.

    Note:
        Beware that pushing to git create a commit and add all the changes, not only the staged ones.

    When using sys args for boolean values, always define True or False.

    There is command line entrypoint called `mypythontools_cicd`. After mypythontools is installed, you can
    use it in terminal like::

        mypythontools_cicd --do_only reformat

    Args:
        config (None | PipelineConfig, optional): It is possible to configure all the params with CLI args
            from terminal. Just create script, where create config, use 'config.with_argparse()' and call
            project_utils_pipeline(config=config). Example usage 'python your_script.py --deploy True'
        do_only (Literal[None, "reformat", "test", "docs", "sync_requirements", "commit_and_push_git", "deploy"], optional):
            Run just single function from pipeline, ignore the others. Reason for why to call it form here and
            not directly is to be able to use sys args or single command line entrypoint. Defaults to None.
        reformat (bool, optional): Reformat all python files with black. Setup parameters in
            `pyproject.toml`, especially setup `line-length`. Defaults to True.
        test (bool, optional): Whether run pytest tests. Defaults to True.
        test_options (None | dict, optional): Parameters of tests function e.g.
            ``{"virtualenvs": ["venv/37", "venv/310], "test_coverage": True, "verbose": False}``.
            Defaults to None.
        version (None | str, optional): New version. E.g. '1.2.5'. If 'increment', than it's auto
            incremented. E.g from '1.0.2' to 'v1.0.3'. If empty string "" or not value arg in CLI,
            then version is not changed. 'Defaults to "increment".
        docs(bool, optional): Whether generate sphinx apidoc and generate rst files for documentation.
            Some files in docs source can be deleted - check `docs` docstrings for details.
            Defaults to True.
        sync_requirements(None | Literal["infer"] | PathLike | Sequence[PathLike], optional): Check
            requirements.txt and update all the libraries. Defaults to False.
        commit_and_push_git (bool, optional): Whether push repository on git with commit_message, tag and tag
            message. Defaults to True.
        commit_message (str, optional): Git message. Defaults to 'New commit'.
        tag (str, optional): Used tag. If tag is '__version__', than updated version from __init__
            is used.  If empty string "" or not value arg in CLI, then tag is not created.
            Defaults to __version__.
        tag_message (str, optional): Tag message. Defaults to New version.
        deploy (bool, optional): Whether deploy to PYPI. `TWINE_USERNAME` and `TWINE_PASSWORD`
            are used for authorization. Defaults to False.
        allowed_branches (None | Sequence[str], optional): As there are stages like pushing to git or to PyPi,
            it's better to secure it to not to be triggered on some feature branch. If not one of
            defined branches, error is raised. Defaults to ("master", "main").
        verbosity (Literal[0, 1, 2], optional): How much information print to console. 0 prints just errors,
            1 prints when starting new step, 2 prints every stdout to console. Defaults to 1.

    Example:
        Recommended use is from IDE (for example with Tasks in VS Code). Check utils docs for how to use it.
        You can also use it from python... ::

            if __name__ == "__main__":
                project_utils_pipeline(commit_and_push_git=False, deploy=False, allowed_branches=None)

        It's also possible to use CLI and configure it via args. This example just push repo to PyPi. ::

            python path-to-project/utils/push_script.py --do_only deploy
    """
    if not config:
        config = PipelineConfig()
        config.update(
            {
                "do_only": do_only,
                "reformat": reformat,
                "test": test,
                "test_options": test_options,
                "version": version,
                "sync_requirements": sync_requirements,
                "docs": docs,
                "commit_and_push_git": commit_and_push_git,
                "commit_message": commit_message,
                "tag": tag,
                "tag_message": tag_message,
                "deploy": deploy,
                "allowed_branches": allowed_branches,
                "verbosity": verbosity,
            }
        )

    if not GLOBAL_VARS.is_tested:
        config.with_argparse()

    if config.do_only:
        config.update(
            {
                "reformat": False,
                "test": False,
                "docs": False,
                "sync_requirements": None,
                "commit_and_push_git": False,
                "deploy": False,
                "version": None,
            }
        )
        config.update({config.do_only: True})

        if config.verbosity == 1:
            config.verbosity = 0

    verbose = True if config.verbosity == 2 else False
    progress_is_printed = config.verbosity > 0

    if config.allowed_branches:
        import git.repo
        from git.exc import InvalidGitRepositoryError

        validate_sequence(allowed_branches, "allowed_branches")

        try:
            branch = git.repo.Repo(PROJECT_PATHS.root.as_posix()).active_branch.name
        except InvalidGitRepositoryError:
            raise RuntimeError(
                "Loading of git project failed. Verify whether running pipeline from correct path. If "
                "checks branch with `allowed_branches', there has to be `.git` folder available."
            ) from None

        if branch not in config.allowed_branches:
            raise RuntimeError(
                "Pipeline started on branch that is not allowed."
                "If you want to use it anyway, add it to allowed_branches parameter and "
                "turn off changing version and creating tag."
            )

    # Do some checks before run pipeline so not need to rollback eventually
    if config.deploy:
        usr = os.environ.get("TWINE_USERNAME")
        pas = os.environ.get("TWINE_PASSWORD")

        if not usr or not pas:
            raise KeyError("Setup env vars TWINE_USERNAME and TWINE_PASSWORD to use deploy.")

    if config.sync_requirements:
        print_progress("Syncing requirements", progress_is_printed)

        if not venvs.is_venv:
            raise RuntimeError("'sync_requirements' available only if using virtualenv.")
        my_venv = venvs.Venv(sys.prefix)
        my_venv.create()
        my_venv.sync_requirements(config.sync_requirements, verbose=verbose)

    if config.test:
        print_progress("Testing", progress_is_printed)

        if not config.test_options:
            config.test_options = {}

        tests.run_tests(**config.test_options, verbosity=verbosity)

    if config.reformat:
        print_progress("Reformatting", progress_is_printed)
        reformat_with_black()

    if config.version and config.version != "None":
        print_progress("Setting version", progress_is_printed)
        original_version = get_version()
        set_version(config.version)

    try:
        if config.docs:
            print_progress("Sphinx docs generation", progress_is_printed)
            docs_regenerate(verbose=verbose)

        if config.commit_and_push_git:
            print_progress("Pushing to github", progress_is_printed)
            git_push(
                commit_message=config.commit_message,
                tag=config.tag,
                tag_message=config.tag_message,
                verbose=verbose,
            )

    except Exception as err:  # pylint: disable=broad-except
        if config.version:
            set_version(original_version)  # type: ignore

        raise RuntimeError(
            f"{3 * EMOJIS.DISAPPOINTMENT} Utils pipeline failed {3 * EMOJIS.DISAPPOINTMENT} \n\n"
            "Original version restored. Nothing was pushed to repo, you can restart pipeline."
        ) from err

    try:
        if config.deploy:
            print_progress("Deploying to PyPi", progress_is_printed)
            deploy_to_pypi()

    except Exception as err:  # pylint: disable=broad-except
        raise RuntimeError(
            f"{3 * EMOJIS.DISAPPOINTMENT} Deploy failed {3 * EMOJIS.DISAPPOINTMENT} \n\n"
            "Already pushed to repository. Deploy manually. Version already changed.",
        ) from err

    print_progress(f"{3 * EMOJIS.PARTY} Finished {3 * EMOJIS.PARTY}", True)



"""
This module build the app via pyinstaller.
It has presets to build applications build with eel.

There is one main function `build_app`. Check it's help for how to use
it (should be very simple).

Note:
    You can run build for example from vs code tasks, create folder utils,
    create build_script.py inside, add

>>> import mypythontools
...
>>> if __name__ == "__main__":
...     mypythontools.build.build_app()  # With all the params you need.

In VS Code you can add this task to global tasks.json to run it with anytime::

    {
        "label": "Build app",
        "type": "shell",
        "command": "python",
        "args": ["${workspaceFolder}/utils/build_script.py"],
        "presentation": {
            "reveal": "always",
            "panel": "new"
        }
    },
"""
from mypythontools_cicd.build.build_internal import build_app

__all__ = ["build_app"]



"""Module with functions for project_utils_functions subpackage."""

from __future__ import annotations
from typing import Sequence
import ast
import re

from mypythontools.paths import validate_path, PathLike
from mypythontools.types import validate_sequence

from mypythontools.misc import delete_files
from mypythontools.system import check_script_is_available, get_console_str_with_quotes, terminal_do_command

# Lazy loaded
# from git import Repo

from mypythontools_cicd.project_paths import PROJECT_PATHS


def reformat_with_black(
    root_path: None | PathLike = None, extra_args: Sequence[str] = ("--quiet",), verbose: bool = False
) -> None:
    """Reformat code with black.

    Args:
        root_path (None | PathLike, optional): Root path of project. If None, will be inferred.
            Defaults to None.
        extra_args (Sequence[str], optional): Some extra args for black. Defaults to ("--quiet,").
        verbose (bool, optional): If True, result of terminal command will be printed to console.
            Defaults to False.

    Example:
        >>> reformat_with_black()
    """
    check_script_is_available("black", "black")
    validate_sequence(extra_args, "extra_args")

    root_path = validate_path(root_path) if root_path else PROJECT_PATHS.root

    terminal_do_command(
        f"black . {' '.join(extra_args)}", cwd=root_path, verbose=verbose, error_header="Formatting failed"
    )


def git_push(
    commit_message: str, tag: str = "__version__", tag_message: str = "New version", verbose: bool = False
) -> None:
    """Stage all changes, commit, add tag and push.

    If tag is `__version__`, then tag is inferred from `__init__.py`.

    Args:
        commit_message (str): Commit message.
        tag (str, optional): Define tag used in push. If tag is '__version__', than is automatically generated
            from __init__ version. E.g from '1.0.2' to 'v1.0.2'.  Defaults to '__version__'.
        tag_message (str, optional): Message in annotated tag. Defaults to 'New version'.
        verbose (bool, optional): If True, result of terminal command will be printed to console.
            Defaults to False.
    """
    import git.repo
    import git.exc

    git_command = f"git add . && git commit -m {get_console_str_with_quotes(commit_message)} && git push"

    if tag == "__version__":
        tag = f"v{get_version()}"

    if tag:
        if not tag_message:
            tag_message = "New version"
        try:
            git.repo.Repo(PROJECT_PATHS.root.as_posix()).create_tag(tag, message=tag_message)
        except git.exc.GitCommandError as err:
            raise RuntimeError("Tag creation failed. It can be because such a tag already exists.") from err

        git_command += " --follow-tags"

    try:
        terminal_do_command(git_command, cwd=PROJECT_PATHS.root.as_posix(), verbose=verbose)

    except RuntimeError as err:
        git.repo.Repo(PROJECT_PATHS.root.as_posix()).delete_tag(tag)  # type: ignore
        raise RuntimeError("Push to git failed. Version restored and created git tag deleted.") from err


def validate_version(version: str):
    """Check whether parsed version is valid.

    Args:
        version (str): E.g "1.0.1"

    Returns:
        bool: Whether is valid.
    """
    return version == "increment" or (
        len(version.split(".")) == 3 and all([i.isdecimal() for i in version.split(".")])
    )


def set_version(
    version: str = "increment",
    init_path: None | PathLike = None,
) -> None:
    """Change your version in your `__init__.py` file.

    Args:
        version (str, optional): Form that is used in `__init__`, so for example "1.2.3". Do not use 'v'
            appendix. If version is 'increment', it will increment your `__version__` in you `__init__.py` by
            0.0.1. Defaults to "increment".
        init_path (None | PathLike, optional): Path of file where `__version__` is defined.
            Usually `__init__.py`. If None, will be inferred. Defaults to None.

    Raises:
        ValueError: If no `__version__` is find.
    """
    init_path = validate_path(init_path) if init_path else PROJECT_PATHS.init

    is_valid = validate_version(version)

    if not is_valid:
        raise ValueError(
            "Version not validated. Version has to be of form '1.2.3'. Three digits and two dots. "
            f"You used {version}"
        )

    with open(init_path.as_posix(), "r") as init_file:

        list_of_lines = init_file.readlines()

        found = False

        for i, j in enumerate(list_of_lines):
            if j.startswith("__version__"):

                found = True

                delimiter = '"' if '"' in j else "'"
                delimited = j.split(delimiter)

                if version == "increment":
                    version_list = delimited[1].split(".")
                    version_list[2] = str(int(version_list[2]) + 1)
                    delimited[1] = ".".join(version_list)

                else:
                    delimited[1] = version

                list_of_lines[i] = delimiter.join(delimited)
                break

        if not found:
            raise ValueError("__version__ variable not found in __init__.py. Try set init.")

    with open(init_path.as_posix(), "w") as init_file:

        init_file.writelines(list_of_lines)


def get_version(init_path: None | PathLike = None) -> str:
    """Get version info from `__init__.py` file.

    Args:
        init_path (None | PathLike, optional): Path to `__init__.py` file. If None, will be inferred.
            Defaults to None.

    Returns:
        str: String of version from `__init__.py`.

    Raises:
        ValueError: If no `__version__` is find. Try set init_path...

    Example:
        >>> version = get_version()
        >>> len(version.split(".")) == 3 and all([i.isdecimal() for i in version.split(".")])
        True
    """
    init_path = validate_path(init_path) if init_path else PROJECT_PATHS.init

    with open(init_path.as_posix(), "r") as init_file:
        version = re.findall('__version__ = "(.*)"', init_file.read())[0]

    if validate_version(version):
        return version
    else:
        raise RuntimeError("Version not found in __init__.py")


def docs_regenerate(
    docs_path: None | PathLike = None,
    build_locally: bool = True,
    git_add: bool = True,
    keep: Sequence[PathLike] = (),
    ignore: Sequence[PathLike] = ("modules.rst", "**/*internal.py"),
    verbose: bool = False,
) -> None:
    """Generate all rst files necessary for sphinx documentation generation with sphinx-apidoc.

    It automatically delete rst files from removed or renamed files.

    Note:
        All the files except ['conf.py', 'index.rst', '_static', '_templates', 'content/**'], and files in
        'keep' parameter will be deleted!!! Because if some files would be deleted or
        renamed, rst would stay and html was generated. If you have some extra files or folders in docs
        source, add it to content folder or to 'keep' parameter.

    Function suppose sphinx build and source in separate folders...

    Args:
        docs_path (None | PathLike, optional): Where source folder is. If None, will be inferred.
            Defaults to None.
        build_locally (bool, optional): If true, build folder with html files locally.
            Defaults to True.
        git_add (bool, optional): Whether to add generated files to stage. False mostly for
            testing reasons. Defaults to True.
        keep (Sequence[PathLike], optional): List of files and folder names that will not be
            deleted. Deletion is because if some file would be renamed or deleted, rst docs would still stay.
            Glob-style patterns can be used, but it's not recursive, but only first level of source folder is
            used. Defaults to None.
        ignore (Sequence[PathLike], optional): Whether ignore some files from generated rst files. For example
            It can be python modules that will be ignored or it can be rst files created, that will be
            deleted. to have no errors in sphinx build for unused modules, or for internal modules. Glob-style
            patterns can be used. Defaults to ("modules.rst", "**/*_.py")
        verbose (bool, optional): If True, result of terminal command will be printed to console.
            Defaults to False.

    Note:
        Function suppose structure of docs like::

            -- docs
            -- -- source
            -- -- -- conf.py
            -- -- make.bat
    """
    check_script_is_available("sphinx-apidoc", "sphinx")

    validate_sequence(keep, keep)
    validate_sequence(ignore, ignore)

    docs_path = validate_path(docs_path) if docs_path else PROJECT_PATHS.docs
    docs_source_path = docs_path / "source"

    source_path = PROJECT_PATHS.app
    source_console_path = get_console_str_with_quotes(source_path)

    keep = [
        *keep,
        "conf.py",
        "index.rst",
        "_static",
        "_templates",
        "content",
    ]
    ignore_list = [*ignore]

    ignored = " "

    for i in ignore_list:
        for file in source_path.rglob(str(i)):
            ignored = ignored + f"{get_console_str_with_quotes(file)} "

    for file in docs_source_path.iterdir():
        if not any((file.match(str(pattern)) for pattern in keep)):
            delete_files(file)

    apidoc_command = (
        f"sphinx-apidoc --module-first --force --separate -o source {source_console_path} {ignored}"
    )

    terminal_do_command(
        apidoc_command, cwd=docs_path, verbose=verbose, error_header="Docs sphinx-apidoc failed."
    )

    if ignore_list:
        for file in docs_source_path.iterdir():
            if any((file.match(str(pattern)) for pattern in ignore_list)):
                delete_files(file)

    if build_locally:
        delete_files
        terminal_do_command(
            "make html", cwd=docs_path, verbose=verbose, error_header="Sphinx build failed.", shell=True
        )

    if git_add:
        terminal_do_command("git add docs", cwd=PROJECT_PATHS.root.as_posix(), verbose=verbose)


def generate_readme_from_init(git_add: bool = True) -> None:
    """Generate README file from `__init__.py` docstrings.

    Because i had very similar things in main `__init__.py` and in readme. It was to maintain news
    in code. For better simplicity i prefer write docs once and then generate. One code, two use cases.

    Why `__init__`? - Because in IDE on mouseover developers can see help.
    Why README.md? - Good for github.com

    Args:
        git_add (bool, optional): Whether to add generated files to stage. False mostly
            for testing reasons. Defaults to True.
    """
    with open(PROJECT_PATHS.init.as_posix()) as init_file:
        file_contents = init_file.read()
    module = ast.parse(file_contents)
    docstrings = ast.get_docstring(module)

    if docstrings is None:
        docstrings = ""

    with open(PROJECT_PATHS.readme.as_posix(), "w") as file:
        file.write(docstrings)

    if git_add:
        terminal_do_command(f"git add {PROJECT_PATHS.readme}", cwd=PROJECT_PATHS.root.as_posix())



"""Module with functions around testing. You can run tests including doctest with generating coverage,
you can generate tests from readme or you can configure tests in conftest with single call."""

from mypythontools_cicd.tests.tests_internal import (
    add_readme_tests,
    deactivate_test_settings,
    run_tests,
    setup_tests,
)

__all__ = ["add_readme_tests", "deactivate_test_settings", "run_tests", "setup_tests"]



"""Module with functions for 'build' subpackage."""
from __future__ import annotations
from typing import Sequence
import shutil
from pathlib import Path
import sys

from typing_extensions import Literal
from mypythontools.paths import PathLike, find_path, validate_path
from mypythontools.misc import delete_files
from mypythontools.system import (
    check_script_is_available,
    get_console_str_with_quotes,
    terminal_do_command,
)

import mylogging

from ..project_paths import PROJECT_PATHS
from .. import venvs

# Lazy imports
# import EelForkExcludeFiles


def build_app(
    root_path: None | str | Path = None,
    main_file: str | Path = "app.py",
    preset: Literal["eel", None] = None,
    web_path: None | str | Path = None,
    build_web: bool | str = "preset",
    virtualenv: None | PathLike = sys.prefix,
    sync_requirements: None | Literal["infer"] | PathLike | Sequence[PathLike] = "infer",
    remove_last_build: bool = False,
    console: bool = True,
    debug: bool = False,
    icon: str | Path | None = None,
    hidden_imports: None | list[str] = None,
    ignored_packages: None | list[str] = None,
    datas: None | list[tuple[str, str]] = None,
    env_vars: None | dict = None,
    name: None | str = None,
    clean: bool = True,
    verbosity: Literal[0, 1, 2] = 1,
) -> None:
    """One script to build .exe app from source code.

    This script automatically generate .spec file, build node web files and add environment variables during
    build.

    This script suppose some structure of the app (may have way different though). You can use project-starter
    from the same repository, if you start with application.

    Args:
        root_path (None | str | Path, optional): Path of root folder where build and dist folders will be
            placed. If None, path is inferred. Defaults to None.
        main_file (str, optional): Main file path or name with extension. Main file is found automatically
            and don't have to be in root. Defaults to 'app.py'.
        preset (Literal['eel', None], optional): Edit other params for specific use cases (append to
            hidden_imports, datas etc.). Defaults to None.
        web_path (None | str | Path, optional): If using eel preset it's folder with index.html. If None,
            path is inferred. Defaults to None.
        build_web (bool | str, optional): If application contain package.json build node application.
            If 'preset' build automatically depending on preset. Defaults to 'preset'.
        virtualenv (None | PathLike, optional): Whether run in virtualenv. If you want to use current venv,
            use ``sys.prefix``. Defaults to sys.prefix.
        sync_requirements (None | Literal["infer"] | PathLike | Sequence[PathLike], optional): If using
            `virtualenv` define what libraries will be installed by path to requirements.txt. Can also be a
            list of more files e.g `["requirements.txt", "requirements_dev.txt"]`. If "infer", autodetected
            (all requirements). Defaults to "infer".
        remove_last_build (bool, optional): If some problems, it is possible to delete build and dist folders.
            Defaults to False.
        console (bool, optional): Before app run terminal window appears (good for debugging).
            Defaults to False.
        debug (bool, optional): If no console, then dialog window with traceback appears.
            Defaults to False.
        icon (None | str | Path, optional): Path or name with extension to .ico file (!no png!).
            Defaults to None.
        hidden_imports (None | list, optional): If app is not working, it can be because some library was
            not builded. Add such libraries into this list. Defaults to None.
        ignored_packages (None | list, optional): Libraries take space even if not necessary.
            Defaults to None.
        datas (None | list[tuple[str, str], optional): Add static files to build.
            Example: [('my_source_path, 'destination_path')]. Defaults to None.
        env_vars (None | dict, optional): Add some env vars during build. Mostly to tell main script that
            it's production (ne development) mode. Defaults to None.
        name (None | str, optional): If name of app is different than main py file. Defaults to None.
        clean (bool, optional): Remove spec file and var env py hook. Defaults to True.
        verbosity (Literal[0, 1, 2], optional): How much information print to console. 0 prints just errors,
            1 prints when starting new step, 2 prints every stdout to console. Defaults to 1.

    Note:
        Build pyinstaller bootloader on your pc, otherwise antivirus can check the
        file for a while on first run and even alert false positive.

        Download from github, cd to bootloader and::

            python ./waf all

        Back to pyinstaller folder and python `setup.py`
    """
    # if sys.version_info.major == 3 and sys.version_info.minor >= 10:
    #     raise RuntimeError(mylogging.format_str("Python version >=3.10 not supported yet."))

    check_script_is_available("pyinstaller")

    verbose = True if verbosity == 2 else False

    if not hidden_imports:
        hidden_imports = []

    if not ignored_packages:
        ignored_packages = []

    if not datas:
        datas = []

    if not env_vars:
        env_vars = {}

    root_path = validate_path(root_path) if root_path else PROJECT_PATHS.root

    # Try to recognize the structure of app
    build_path = root_path / "build"

    if not build_path.exists():
        build_path.mkdir(parents=True, exist_ok=True)

    # Remove last dist manually to avoid permission error if opened in some application
    dist_path = root_path / "dist"

    if dist_path.exists():
        try:
            shutil.rmtree(dist_path, ignore_errors=False)
        except (PermissionError, OSError) as remove_exception:

            raise PermissionError(
                mylogging.format_str(
                    "App is opened (May be in another app(terminal, explorer...)). Close it first."
                )
            ) from remove_exception

    # May be just name - not absolute
    main_file_path = Path(main_file)

    if not main_file_path.exists():

        # Iter paths and find the one
        main_file_path = find_path(
            main_file_path.name,
        )

        if not main_file_path.exists():
            raise KeyError("Main file not found, not inferred and must be configured in params...")

    main_file_path = main_file_path.resolve()

    if not name:
        name = main_file_path.stem

    main_folder_path = main_file_path.parent

    if icon:
        icon_path = Path(icon)

        if not icon_path.exists():

            # Iter paths and find the one
            icon_path = find_path(
                icon_path.name,
                exclude_names=["node_modules", "build"],
            )

            if not icon_path.exists():
                raise KeyError("Icon not found, not inferred check path or name...")

            icon_path = f"'{icon_path}'"
    else:
        icon_path = None

    generated_warning = """
#########################
### File is generated ###
#########################

# Do not edit this file, edit build_script
"""

    if remove_last_build:
        shutil.rmtree("build", ignore_errors=True)

    # Build JS to static asset
    if build_web is True or (build_web == "preset" and preset in ["eel"]):
        if verbosity:
            print("\n\nBuilding web.\n\n")

        gui_path = find_path("package.json").parent

        check_script_is_available(
            "npm", message="If building application with pyvueel, Node has to be installed."
        )

        terminal_do_command(
            "npm run build",
            cwd=gui_path.as_posix(),
            shell=True,
            verbose=verbose,
            error_header="Build of web files failed.",
        )

    if build_web or preset == "eel":
        if not web_path:
            web_path = find_path(
                "index.html",
                exclude_names=[
                    "public",
                    "node_modules",
                    "build",
                ],
            ).parent

        else:
            web_path = Path(web_path)

        if not web_path.exists():
            raise KeyError("Build web assets not found, not inferred and must be configured in params...")

        datas = [
            *datas,
            (web_path.as_posix(), "gui"),
        ]

    if preset == "eel":

        import EelForkExcludeFiles

        hidden_imports = [
            *hidden_imports,
            "EelForkExcludeFiles",
            "bottle_websocket",
        ]
        datas = [
            *datas,
            (
                EelForkExcludeFiles._eel_js_file,  # pylint: disable=protected-access
                "EelForkExcludeFiles",
            ),
        ]
        env_vars = {
            **env_vars,
            "MY_PYTHON_VUE_ENVIRONMENT": "production",
        }

    if env_vars:
        env_vars_template = f"""
{generated_warning}

import os
for i, j in {env_vars}.items():
    os.environ[i] = j
"""

        env_path = build_path / "env_vars.py"

        with open(env_path, "w") as env_vars_py:
            env_vars_py.write(env_vars_template)
        runtime_hooks = [env_path.as_posix()]
    else:
        env_path = None
        runtime_hooks = None

    spec_template = f"""
{generated_warning}

import sys
from pathlib import Path
import os

sys.setrecursionlimit(5000)
block_cipher = None

a = Analysis(['{main_file_path.as_posix()}'],
            pathex=['{main_folder_path.as_posix()}'],
            binaries=[],
            datas={datas},
            hiddenimports={hidden_imports},
            hookspath=[],
            runtime_hooks={runtime_hooks},
            excludes={ignored_packages},
            win_no_prefer_redirects=False,
            win_private_assemblies=False,
            cipher=block_cipher,
            noarchive=False)
pyz = PYZ(a.pure, a.zipped_data,
            cipher=block_cipher)
exe = EXE(pyz,
        a.scripts,
        [],
        exclude_binaries=True,
        name='{name}',
        debug={debug},
        bootloader_ignore_signals=False,
        strip=False,
        upx=True,
        console={console},
        icon={icon_path})
coll = COLLECT(exe,
            a.binaries,
            a.zipfiles,
            a.datas,
            strip=False,
            upx=True,
            upx_exclude=[],
            name='{name}')
"""
    spec_path = build_path / "app.spec"

    with open(spec_path, "w") as spec_file:
        spec_file.write(spec_template)

    spec_path = get_console_str_with_quotes(spec_path.as_posix())

    if virtualenv:
        my_venv = venvs.Venv(virtualenv)
        my_venv.create()
        if sync_requirements:
            my_venv.sync_requirements(sync_requirements)

        pyinstaller_script = get_console_str_with_quotes(my_venv.scripts_path / "pyinstaller.exe")
        command = my_venv.activate_command + " && " + pyinstaller_script + " --noconfirm " + spec_path

    else:
        command = f"pyinstaller --noconfirm {spec_path}"

    terminal_do_command(
        command,
        cwd=PROJECT_PATHS.root.as_posix(),
        shell=True,
        verbose=verbose,
        error_header="Build with pyinstaller failed. Troubleshooting: Try to install newset pyinstaller locally with "
        "'python setup.py install', update setuptools, delete 'build' and 'dist' folder and try again.",
    )

    if clean:
        to_delete = [spec_path] if not env_path else [spec_path, env_path]
        delete_files(to_delete)



"""
Allow to deploy project.

Possible destinations:

 - PyPi

Check `deploy_to_pypi` function docs for how to use it.

Usually this function is not called manually, but it's a part of `project_utils_pipeline` from cicd. Check it's
docs where it is described, how to use VS Code Task to be able to optionally test, push and deploy with tasks
(one button click).
"""
from mypythontools_cicd.deploy.deploy_internal import deploy_to_pypi

__all__ = ["deploy_to_pypi"]



"""Functions for project utility like docs generation, reformatting etc.

This module contains functions that are usually called in pipeline with 'project_utils_pipeline'
from 'project_utils' which you can call with script from terminal and you can use sys args to configure it.

You can use functions separately of course.
"""

from mypythontools_cicd.project_utils.project_utils_functions.project_utils_functions_internal import (
    generate_readme_from_init,
    get_version,
    git_push,
    reformat_with_black,
    set_version,
    docs_regenerate,
)

__all__ = [
    "generate_readme_from_init",
    "get_version",
    "git_push",
    "reformat_with_black",
    "set_version",
    "docs_regenerate",
]
